import logging
from http import HTTPStatus
from logging import Logger
from typing import List, Union

import click
import uvicorn
from fastapi import FastAPI, Request, Security
from fastapi.middleware.cors import CORSMiddleware
from fastapi.routing import APIRouter
from injector import inject
from opencensus.ext.azure.trace_exporter import AzureExporter
from opencensus.trace.attributes_helper import COMMON_ATTRIBUTES
from opencensus.trace.samplers import ProbabilitySampler
from opencensus.trace.span import SpanKind
from opencensus.trace.tracer import Tracer
from pydantic import AnyHttpUrl

from isar.apis.models.models import ControlMissionResponse, StartMissionResponse
from isar.apis.robot_control.robot_controller import RobotController
from isar.apis.schedule.scheduling_controller import SchedulingController
from isar.apis.security.authentication import Authenticator
from isar.config.configuration_error import ConfigurationError
from isar.config.keyvault.keyvault_error import KeyvaultError
from isar.config.keyvault.keyvault_service import Keyvault
from isar.config.settings import settings

HTTP_URL = COMMON_ATTRIBUTES["HTTP_URL"]
HTTP_STATUS_CODE = COMMON_ATTRIBUTES["HTTP_STATUS_CODE"]


class API:
    @inject
    def __init__(
        self,
        authenticator: Authenticator,
        scheduling_controller: SchedulingController,
        robot_controller: RobotController,
        keyvault_client: Keyvault,
        port: int = settings.API_PORT,
        azure_ai_logging_enabled: bool = settings.LOG_HANDLER_APPLICATION_INSIGHTS_ENABLED,
    ) -> None:
        self.authenticator: Authenticator = authenticator
        self.scheduling_controller: SchedulingController = scheduling_controller
        self.robot_controller: RobotController = robot_controller
        self.keyvault_client: Keyvault = keyvault_client
        self.host: str = "0.0.0.0"  # Locking uvicorn to use 0.0.0.0
        self.port: int = port
        self.azure_ai_logging_enabled: bool = azure_ai_logging_enabled

        self.logger: Logger = logging.getLogger("api")

        self.app: FastAPI = self._create_app()

    def get_app(self) -> FastAPI:
        return self.app

    def run_app(self) -> None:
        uvicorn.run(
            self.app,
            port=self.port,
            host=self.host,
            reload=False,
            log_config=None,
        )

    def _create_app(self) -> FastAPI:
        tags_metadata = [
            {
                "name": "Scheduler",
                "description": "Mission functionality",
            }
        ]
        app = FastAPI(
            openapi_tags=tags_metadata,
            on_startup=[self.authenticator.load_config, self._log_startup_message],
            swagger_ui_oauth2_redirect_url="/oauth2-redirect",
            swagger_ui_init_oauth={
                "usePkceWithAuthorizationCodeGrant": True,
                "clientId": settings.AZURE_CLIENT_ID,
            },
        )

        if self.authenticator.should_authenticate():
            backend_cors_origins: List[Union[str, AnyHttpUrl]] = [
                f"http://{self.host}:{self.port}"
            ]

            app.add_middleware(
                CORSMiddleware,
                allow_origins=[str(origin) for origin in backend_cors_origins],
                allow_credentials=True,
                allow_methods=["*"],
                allow_headers=["*"],
            )

        if self.azure_ai_logging_enabled:
            self._add_request_logging_middleware(app)

        app.include_router(router=self._create_scheduler_router())

        app.include_router(router=self._create_info_router())

        app.include_router(router=self._create_media_control_router())

        return app

    def _create_scheduler_router(self) -> APIRouter:
        router: APIRouter = APIRouter(tags=["Scheduler"])

        authentication_dependency: Security = Security(self.authenticator.get_scheme())

        router.add_api_route(
            path="/schedule/start-mission/{id}",
            endpoint=self.scheduling_controller.start_mission_by_id,
            methods=["POST"],
            deprecated=True,
            dependencies=[authentication_dependency],
            summary="Start a mission with id='id' from the current mission planner",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully started",
                    "model": StartMissionResponse,
                },
                HTTPStatus.NOT_FOUND.value: {
                    "description": "Not found - Mission not found",
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            path="/schedule/start-mission",
            endpoint=self.scheduling_controller.start_mission,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Start the mission provided in JSON format",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully started",
                    "model": StartMissionResponse,
                },
                HTTPStatus.UNPROCESSABLE_ENTITY.value: {
                    "description": "Invalid body - The JSON is incorrect",
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            path="/schedule/stop-mission",
            endpoint=self.scheduling_controller.stop_mission,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Stop the current mission",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully stopped",
                    "model": ControlMissionResponse,
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            path="/schedule/pause-mission",
            endpoint=self.scheduling_controller.pause_mission,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Pause the current mission",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully paused",
                    "model": ControlMissionResponse,
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            path="/schedule/resume-mission",
            endpoint=self.scheduling_controller.resume_mission,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Resume the currently paused mission - if any",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully resumed",
                    "model": ControlMissionResponse,
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            path="/schedule/drive-to",
            endpoint=self.scheduling_controller.drive_to,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Drive to the provided pose",
            deprecated=True,
            responses={
                HTTPStatus.OK.value: {
                    "description": "Drive to succesfully started",
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            path="/schedule/start-localization-mission",
            endpoint=self.scheduling_controller.start_localization_mission,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Localize at the provided pose",
            deprecated=True,
            responses={
                HTTPStatus.OK.value: {
                    "description": "Localization succesfully started",
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            path="/schedule/move_arm/{arm_pose_literal}",
            endpoint=self.scheduling_controller.start_move_arm_mission,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Move arm to the given arm pose literal",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Move arm mission successfully started",
                },
                HTTPStatus.BAD_REQUEST.value: {
                    "description": "A move arm mission was scheduled on a robot that "
                    "does not support it or the input was incorrect",
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )

        return router

    def _create_info_router(self) -> APIRouter:
        router: APIRouter = APIRouter(tags=["Info"])

        authentication_dependency: Security = Security(self.authenticator.get_scheme())

        router.add_api_route(
            path="/info/robot-settings",
            endpoint=self.robot_controller.get_info,
            methods=["GET"],
            dependencies=[authentication_dependency],
            summary="Information about the robot-settings",
        )

        return router

    def _create_media_control_router(self) -> APIRouter:
        router: APIRouter = APIRouter(tags=["Media"])

        authentication_dependency: Security = Security(self.authenticator.get_scheme())

        router.add_api_route(
            path="/media/media-stream-config",
            endpoint=self.robot_controller.generate_media_config,
            methods=["GET"],
            dependencies=[authentication_dependency],
            summary="Generates a media stream connection config",
        )

        return router

    def _log_startup_message(self) -> None:
        address_format = "%s://%s:%d/docs"
        message = f"Uvicorn running on {address_format} (Press CTRL+C to quit)"
        protocol = "http"
        color_message = (
            "Uvicorn running on "
            + click.style(address_format, bold=True)
            + " (Press CTRL+C to quit)"
        )
        self.logger.info(
            message,
            protocol,
            self.host,
            self.port,
            extra={"color_message": color_message},
        )

    def _add_request_logging_middleware(self, app: FastAPI) -> None:
        connection_string: str
        try:
            connection_string = self.keyvault_client.get_secret(
                "application-insights-connection-string"
            ).value
        except KeyvaultError:
            message: str = (
                "Missing connection string for Application Insights in key vault. "
            )
            self.logger.critical(message)
            raise ConfigurationError(message)

        @app.middleware("http")
        async def middlewareOpencensus(request: Request, call_next):
            tracer = Tracer(
                exporter=AzureExporter(connection_string=connection_string),
                sampler=ProbabilitySampler(1.0),
            )
            with tracer.span("main") as span:
                span.span_kind = SpanKind.SERVER

                response = await call_next(request)

                tracer.add_attribute_to_current_span(
                    attribute_key=HTTP_STATUS_CODE, attribute_value=response.status_code
                )
                tracer.add_attribute_to_current_span(
                    attribute_key=HTTP_URL, attribute_value=str(request.url)
                )

            return response



from abc import ABCMeta, abstractmethod

from robot_interface.models.mission.mission import Mission


class MissionPlannerInterface(metaclass=ABCMeta):
    @abstractmethod
    def get_mission(self, mission_id: str) -> Mission:
        """
        Parameters
        ----------
        mission_id : str

        Returns
        -------
        mission : Mission
        """
        pass


class MissionPlannerError(Exception):
    pass


class MissionNotFoundError(Exception):
    pass



from typing import Any

from robot_interface.telemetry.mqtt_client import MqttClientInterface


class MqttClientMock(MqttClientInterface):
    def publish(
        self, topic: str, payload: Any, qos: int = 0, retain: bool = False
    ) -> None:
        pass



from typing import Iterator, List

from isar.mission_planner.task_selector_interface import (
    TaskSelectorInterface,
    TaskSelectorStop,
)
from robot_interface.models.mission.task import TASKS


class SequentialTaskSelector(TaskSelectorInterface):
    def __init__(self) -> None:
        super().__init__()
        self._iterator: Iterator = None

    def initialize(self, tasks: List[TASKS]) -> None:
        super().initialize(tasks=tasks)
        self._iterator = iter(self.tasks)

    def next_task(self) -> TASKS:
        try:
            return next(self._iterator)
        except StopIteration:
            raise TaskSelectorStop



from dataclasses import dataclass
from enum import Enum
from typing import Optional


class ErrorReason(str, Enum):
    RobotCommunicationException = "robot_communication_exception"
    RobotCommunicationTimeoutException = "robot_communication_timeout_exception"
    RobotInfeasibleTaskException = "robot_infeasible_task_exception"
    RobotInfeasibleMissionException = "robot_infeasible_mission_exception"
    RobotMissionStatusException = "robot_mission_status_exception"
    RobotTaskStatusException = "robot_task_status_exception"
    RobotAPIException = "robot_api_exception"
    RobotActionException = "robot_action_exception"
    RobotInitializeException = "robot_initialize_exception"
    RobotRetrieveDataException = "robot_retrieve_data_exception"
    RobotRetrieveInspectionException = "robot_retrieve_inspection_exception"
    RobotTelemetryException = "robot_telemetry_exception"
    RobotTelemetryNoUpdateException = "robot_telemetry_no_update_exception"
    RobotTelemetryPoseException = "robot_telemetry_pose_exception"
    RobotMapException = "robot_map_exception"
    RobotTransformException = "robot_transform_exception"
    RobotUnknownErrorException = "robot_unknown_error_exception"
    RobotDisconnectedException = "robot_disconnected_exception"
    RobotMissionNotSupportedException = "robot_mission_not_supported_exception"
    RobotMissionMissingStartPoseException = "robot_mission_missing_start_pose_exception"


@dataclass
class ErrorMessage:
    error_reason: Optional[ErrorReason]
    error_description: str


# This is the base exception class for exceptions that should be raised from the robot
# package and handled in ISAR. Please peruse the different subclasses for information
# on which exceptions to use where.
class RobotException(Exception):
    def __init__(self, error_reason: ErrorReason, error_description: str):
        self.error_reason: ErrorReason = error_reason
        self.error_description: str = error_description


# An exception which should be thrown by the robot package if it is unable to
# communicate with the robot API. ISAR will retry the request.
class RobotCommunicationException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotCommunicationException,
            error_description=error_description,
        )

    pass


# An exception which should be thrown by the robot package if the communication has timed
# out and ISAR should retry the request.
class RobotCommunicationTimeoutException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotCommunicationTimeoutException,
            error_description=error_description,
        )

    pass


# An exception which should be thrown by the robot package if it is unable to start the
# current task.
class RobotInfeasibleTaskException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotInfeasibleTaskException,
            error_description=error_description,
        )

    pass


# An exception which should be thrown by the robot package if it is unable to start the
# current mission.
class RobotInfeasibleMissionException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotInfeasibleMissionException,
            error_description=error_description,
        )

    pass


# An exception which should be thrown by the robot package if it is unable to collect
# the status of the current mission.
class RobotMissionStatusException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotMissionStatusException,
            error_description=error_description,
        )

    pass


# An exception which should be thrown by the robot package if it is unable to collect
# the status of the current task.
class RobotTaskStatusException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotTaskStatusException,
            error_description=error_description,
        )

    pass


# An exception which should be thrown by the robot package if it is able to communicate
# with the robot API but the result of the communication leads to an exception. An
# example could be a KeyError while reading from the response dictionary.
class RobotAPIException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotAPIException,
            error_description=error_description,
        )

    pass


# An exception which should be thrown by the robot package if it is unable to perform a
# requested action. For example the package is unable to stop the robot.
class RobotActionException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotActionException,
            error_description=error_description,
        )

    pass


# An exception which should be thrown by the robot package if something is wrong during
# the initialization of the robot. This exception will cause the mission to fail as
# initialization is performed prior to starting the mission.
class RobotInitializeException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotInitializeException,
            error_description=error_description,
        )

    pass


# An exception which should be thrown by the robot package if it is unable to retrieve
# data from the API like currently executing missions, status of the current mission
# and similar.
class RobotRetrieveDataException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotRetrieveDataException,
            error_description=error_description,
        )

    pass


# An exception which should be thrown by the robot package if it is unable to collect
# the inspections that were generated for the currently executing task or mission.
class RobotRetrieveInspectionException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotRetrieveInspectionException,
            error_description=error_description,
        )

    pass


# An exception which should be thrown by the robot package if it is unable to retrieve
# telemetry data. It should be used exclusively by the telemetry publishers and their
# functions.
class RobotTelemetryException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotTelemetryException,
            error_description=error_description,
        )

    pass


# An exception which should be thrown by the robot package if it is unable to retrieve
# telemetry pose data. It should be used exclusively by the telemetry pose publisher.
class RobotTelemetryPoseException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotTelemetryPoseException,
            error_description=error_description,
        )

    pass


# An exception which should be thrown by the robot package if there is no new telemetry update.
class RobotTelemetryNoUpdateException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotTelemetryNoUpdateException,
            error_description=error_description,
        )

    pass


# An exception which should be thrown by the robot package if it is unable to load the
# configuration for maps and transformations. This could be caused by faulty
# configuration and this exception will cause ISAR to crash as further execution is not
# advised.
class RobotMapException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotMapException,
            error_description=error_description,
        )

    pass


# An exception which should be thrown by the robot package if it is unable to transform
# the coordinates correctly between asset and robot frame.
class RobotTransformException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotTransformException,
            error_description=error_description,
        )

    pass


# An exception which should be thrown by the robot package if something occurred that
# was unexpected and the error reason is unknown.
class RobotUnknownErrorException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotUnknownErrorException,
            error_description=error_description,
        )

    pass


# An exception which should be thrown by the robot package if the robot is not connected to the cloud
class RobotDisconnectedException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotDisconnectedException,
            error_description=error_description,
        )

    pass


# An exception which should be thrown by the robot package if the robot is given a mission type it cannot run, such as a localisation mission
class RobotMissionNotSupportedException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotMissionNotSupportedException,
            error_description=error_description,
        )

    pass


# An exception which should be thrown by the robot package if the mission is missing start pose and it needed it
class RobotMissionMissingStartPoseException(RobotException):
    def __init__(self, error_description: str) -> None:
        super().__init__(
            error_reason=ErrorReason.RobotMissionMissingStartPoseException,
            error_description=error_description,
        )

    pass
